#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h> 
#include <unistd.h> 
#include <pthread.h> 
#include <signal.h>
#include <strings.h>
#include <stdlib.h>
#include <fcntl.h>

#define SERVER_PORT 7111
#define LEN_RECVBUFFER 1000


char textstr[10000];
int sfd;

/*--------------------------------------------------------------------*/

//Signalhandler zum Abbrechen des Servers
void sighandler(int sig)
{
    
    switch( sig ) {
            
        case SIGPIPE: // Nur den zugehörigen Client abbrechen
            break;
            
        default:
            printf("Signal %d caught\n", sig);
            
            if(0 > close(sfd))
            {
                perror("Problem with close()");
                exit(1);
            }
            
            exit(0);
            break;
    }
    
    
}

/*--------------------------------------------------------------------*/

// Thread function:
// Eingabeparameter ist 1 Pointer.
// Falls mehrere Variablen uebergeben werden sollen,
// sird eine Struktur verwendet, auf die der Pointer zeigt.
void *client_handler(void *p) {
    
    int clientsfd = *(int *)p;
    char *recvbuffer = (char *)calloc(LEN_RECVBUFFER+1,1);
    int lenrecv;
    int lensent;
    int len;
    int iCnt;
    int *retval;
    
    fcntl(clientsfd,O_NONBLOCK);
    
    // den dynamisch angelegten Datenbereich hinter p
    // wieder freigeben
    free(p);
    
    while ( 1 ) {
        
        //lesen
        if ( 0 > (lenrecv = 
                  read(clientsfd,
                       recvbuffer,
                       LEN_RECVBUFFER)) ) {
                   }
        
        
        if ( 0 < lenrecv ) {
            recvbuffer[lenrecv] = 0;
            printf("Server-Thread %d received %s\n",clientsfd,recvbuffer);
            
            //wenn STOP empfangen wird, beendet sich der Client-Thread
            if ( !strncmp(recvbuffer,"STOP",strlen("STOP")) ) {
                printf("Server-Child exits\n");
                close(clientsfd);
                free(recvbuffer);
                return(NULL);
            }
            
            for (iCnt=0; iCnt <= strlen(recvbuffer); iCnt++) {
                textstr[iCnt] = recvbuffer[iCnt];
            }
            printf("TEXTSTR: %s\n",textstr);
            
            // wir moechten lenrcv Bytes ueber Socket versenden
            // Dazu machen wir eine Schleife, bis tatsaechlich
            // alle Bytes vom BS akzeptiert wurden.
            // Bei jedem Write() wird die Anfangsadresse des
            // Sendepuffers auf das ERSTE NOCH NICHT VERSENDETE
            // Byte gesetzt. 
            lensent = 0;
            while ( lensent < lenrecv ) {
                if ( 0 > (len = write(clientsfd,
                                      &(recvbuffer[lensent]),
                                      lenrecv - lensent)) ) {
                    perror("write()-call");
                    return NULL;
                }
                lensent = lensent + len;
            }
        }
    }
}

/*--------------------------------------------------------------------*/
int main() {
    int clientsfd;
    struct  sockaddr_in saddr;
    struct sockaddr clientaddr;
    socklen_t lenclientaddr;
    pid_t childpid;
    ssize_t lenrecv, lensent, len;
    int optval;
    
    pthread_t th;
    pthread_attr_t attr;
    int *pPtr;
    
    //Signalhandler zum Beenden
    signal(SIGTERM, sighandler);
    signal(SIGPIPE, sighandler);
    signal(SIGINT, sighandler);
    
    /*--- PF_INET waehlt die IP-Protokollfamilie aus.
     SOCK_STREAM wahlt Byte-Strom-orientierte Kommunikation aus.
     Das bedeutet zZt automatisch TCP ueber IP ---*/
    if ( 0 > ( sfd = socket(PF_INET,SOCK_STREAM,0) ) ) {
        perror("socket()-call");
        exit(1);
    }
    
    /*--- das BS darf die IP/PORT-Adresse sofort
     wieder vergeben: ---*/ 
    optval = 1;
    if ( -1 == setsockopt(sfd,SOL_SOCKET,SO_REUSEADDR,
                          &optval,sizeof(optval)) ) {
        perror("setsockopt()-call");
    }
    
    /* Server bindet IP-Adresse und Port an den Socket ---*/ 
    bzero((char *)&saddr,sizeof(saddr)); // saddr initialisieren
    saddr.sin_family = AF_INET; // Protokollfamilie
    saddr.sin_addr.s_addr = htonl(INADDR_ANY); // trage 
    // IP Adresse des lokalen Rechners ein.
    //htonl(): "Host to Network LONG" (32-Bit-Worte): 
    // IP und Port-Adressen muessen in der NETWORK
    // BYTE ORDER (Big Endian Format) angegeben werden. 
    // htons(): Wandlung von 16-Bit Worten in Network
    // Byte Order.
    saddr.sin_port = htons(SERVER_PORT);
    
    if ( 0 > bind(sfd,(struct sockaddr *)&saddr, sizeof(saddr)) ) {
        perror("bind()-call");
        exit(1);
    }
    
    // Server teilt dem BS mit, dass er nun bereit ist,
    // auf dem Socket zu hoeren. "5" ist das BACKLOG, dh, die
    // Anzahl der Verbindungsaufbau-Wuensche, die das BS 
    // intern puffert.
    if ( 0 > listen(sfd,5) ) {
        perror("listen()-call");
        exit(1);
    }
    
    // Standardverfahren fuer das Entwurfsmuster PARALELL SERVER:
    // Fuer jeden aktiven Client wird vom Server ein Kindprozess
    // erzeugt, der diesen Client bedient.
    while (1) {
        
        lenclientaddr = sizeof(clientaddr);
        // accept() liefert einen NEUEN Socket Filedescriptor
        // zurueck, mit dem die Kommunikation zum CLient,
        // der das zugehoerige connect() ausgeloest hat
        // abgewickelt wird.
        // &clientaddr gibt die Adresse des Puffers an,
        // in die das BS die Adresse des Clients schreiben soll.
        // Bei lenclientaddr muss VOR dem Aufruf die
        // maximale Laenge von clientaddr stehen. NACH dem
        // Aufruf steht in lenclientaddr die tatsaechliche
        // Laenge der empfangenen Adresse.
        if ( 0 > (clientsfd = 
                  accept(sfd,&clientaddr,&lenclientaddr)) ) {
            perror("accept()-call");
            exit(1);
        }
        
        printf("Clientsfd %d\n", clientsfd);
        
        if ( 0 <= clientsfd ) {
            
            // Hier wird ein thread  erzeugt:
            pPtr = (int *)malloc(sizeof(int));
            *pPtr = clientsfd;
            // th: variable fuer Thread Identifikation
            // 2. Parameter: Auswahl der Scheduling
            // Policy; SCHED_FIFO, SCHED_RR oder SCHED_OTHER
            // 3. Parameter: Funktionspointer auf den Thread
            // statt pPtr kann man nicht einfach &clientsfd
            // uebergeben: Zum Zeitpunkt, zu dem der Thread
            // tatsaechlich aktiv, wird, kann in clientsfd
            // schon ein anderer Wert stehen.
            if ( pthread_create(&th,NULL,client_handler,pPtr) ) {
                perror("pthread_create()-call");
                exit(1);
            }
            
        }
        
    }
    
}




